@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -26,8 +26,8 @@
 #include "../../config.hpp"
 #include "../../detail/various.hpp"
 
-#include "../../intrinsics.hpp"
 #include "../../functional.hpp"
+#include "../../intrinsics.hpp"
 
 #include "../../warp/warp_sort.hpp"
 
@@ -36,14 +36,12 @@ BEGIN_ROCPRIM_NAMESPACE
 namespace detail
 {
 
-template<
-    class Key,
-    unsigned int BlockSizeX,
-    unsigned int BlockSizeY,
-    unsigned int BlockSizeZ,
-    unsigned int ItemsPerThread,
-    class Value
->
+template<class Key,
+         unsigned int BlockSizeX,
+         unsigned int BlockSizeY,
+         unsigned int BlockSizeZ,
+         unsigned int ItemsPerThread,
+         class Value>
 class block_sort_bitonic
 {
     static constexpr unsigned int BlockSize     = BlockSizeX * BlockSizeY * BlockSizeZ;
@@ -65,11 +63,9 @@ class block_sort_bitonic
 public:
     using storage_type = detail::raw_storage<storage_type_<Key, Value>>;
 
-    template <class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void sort(Key& thread_key,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    template<class BinaryFunction>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        sort(Key& thread_key, storage_type& storage, BinaryFunction compare_function)
     {
         this->sort_impl<BlockSize, ItemsPerThread>(
             ::rocprim::flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>(),
@@ -79,10 +75,9 @@ public:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void sort(Key (&thread_keys)[ItemsPerThread],
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void sort(Key (&thread_keys)[ItemsPerThread],
+                                            storage_type&  storage,
+                                            BinaryFunction compare_function)
     {
         this->sort_impl<BlockSize, ItemsPerThread>(
             ::rocprim::flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>(),
@@ -92,29 +87,25 @@ public:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
-    void sort(Key& thread_key,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void sort(Key& thread_key, BinaryFunction compare_function)
     {
         ROCPRIM_SHARED_MEMORY storage_type storage;
         this->sort(thread_key, storage, compare_function);
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
-    void sort(Key (&thread_keys)[ItemsPerThread],
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void sort(Key (&thread_keys)[ItemsPerThread],
+                                                  BinaryFunction compare_function)
     {
         ROCPRIM_SHARED_MEMORY storage_type storage;
         this->sort(thread_keys, storage, compare_function);
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void sort(Key& thread_key,
-              Value& thread_value,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void sort(Key&           thread_key,
+                                            Value&         thread_value,
+                                            storage_type&  storage,
+                                            BinaryFunction compare_function)
     {
         this->sort_impl<BlockSize, ItemsPerThread>(
             ::rocprim::flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>(),
@@ -139,20 +130,17 @@ public:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
-    void sort(Key& thread_key,
-              Value& thread_value,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
+        sort(Key& thread_key, Value& thread_value, BinaryFunction compare_function)
     {
         ROCPRIM_SHARED_MEMORY storage_type storage;
         this->sort(thread_key, thread_value, storage, compare_function);
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
-    void sort(Key (&thread_keys)[ItemsPerThread],
-              Value (&thread_values)[ItemsPerThread],
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void sort(Key (&thread_keys)[ItemsPerThread],
+                                                  Value (&thread_values)[ItemsPerThread],
+                                                  BinaryFunction compare_function)
     {
         ROCPRIM_SHARED_MEMORY storage_type storage;
         this->sort(thread_keys, thread_values, storage, compare_function);
@@ -215,42 +203,44 @@ public:
     }
 
 private:
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void copy_to_shared(Key& k, const unsigned int flat_tid, storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        copy_to_shared(Key& k, const unsigned int flat_tid, storage_type& storage)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
-        storage_.key[flat_tid] = k;
+        storage_.key[flat_tid]              = k;
         ::rocprim::syncthreads();
     }
 
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void copy_to_shared(Key (&k)[ItemsPerThread], const unsigned int flat_tid, storage_type& storage) {
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        copy_to_shared(Key (&k)[ItemsPerThread], const unsigned int flat_tid, storage_type& storage)
+    {
         storage_type_<Key, Value>& storage_ = storage.get();
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
             storage_.key[item * BlockSize + flat_tid] = k[item];
         }
         ::rocprim::syncthreads();
     }
 
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void copy_to_shared(Key& k, Value& v, const unsigned int flat_tid, storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        copy_to_shared(Key& k, Value& v, const unsigned int flat_tid, storage_type& storage)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
-        storage_.key[flat_tid] = k;
-        storage_.value[flat_tid] = v;
+        storage_.key[flat_tid]              = k;
+        storage_.value[flat_tid]            = v;
         ::rocprim::syncthreads();
     }
 
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void copy_to_shared(Key (&k)[ItemsPerThread],
-                        Value (&v)[ItemsPerThread],
-                        const unsigned int flat_tid,
-                        storage_type&      storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void copy_to_shared(Key (&k)[ItemsPerThread],
+                                                      Value (&v)[ItemsPerThread],
+                                                      const unsigned int flat_tid,
+                                                      storage_type&      storage)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
             storage_.key[item * BlockSize + flat_tid]   = k[item];
             storage_.value[item * BlockSize + flat_tid] = v[item];
         }
@@ -258,18 +248,18 @@ private:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void swap(Key& key,
-              const unsigned int flat_tid,
-              const unsigned int next_id,
-              const bool dir,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void swap(Key&               key,
+                                            const unsigned int flat_tid,
+                                            const unsigned int next_id,
+                                            const bool         dir,
+                                            storage_type&      storage,
+                                            BinaryFunction     compare_function)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
-        Key next_key = storage_.key[next_id];
-        bool compare = (next_id < flat_tid) ? compare_function(key, next_key) : compare_function(next_key, key);
-        bool swap = compare ^ dir;
+        Key                        next_key = storage_.key[next_id];
+        bool                       compare  = (next_id < flat_tid) ? compare_function(key, next_key)
+                                                                   : compare_function(next_key, key);
+        bool                       swap     = compare ^ dir;
         if(swap)
         {
             key = next_key;
@@ -277,20 +267,21 @@ private:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void swap(Key (&key)[ItemsPerThread],
-              const unsigned int flat_tid,
-              const unsigned int next_id,
-              const bool dir,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void swap(Key (&key)[ItemsPerThread],
+                                            const unsigned int flat_tid,
+                                            const unsigned int next_id,
+                                            const bool         dir,
+                                            storage_type&      storage,
+                                            BinaryFunction     compare_function)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
-            Key next_key = storage_.key[item * BlockSize + next_id];
-            bool compare = (next_id < flat_tid) ? compare_function(key[item], next_key) : compare_function(next_key, key[item]);
-            bool swap = compare ^ dir;
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
+            Key  next_key = storage_.key[item * BlockSize + next_id];
+            bool compare  = (next_id < flat_tid) ? compare_function(key[item], next_key)
+                                                 : compare_function(next_key, key[item]);
+            bool swap     = compare ^ dir;
             if(swap)
             {
                 key[item] = next_key;
@@ -299,44 +290,43 @@ private:
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void swap(Key& key,
-              Value& value,
-              const unsigned int flat_tid,
-              const unsigned int next_id,
-              const bool dir,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void swap(Key&               key,
+                                            Value&             value,
+                                            const unsigned int flat_tid,
+                                            const unsigned int next_id,
+                                            const bool         dir,
+                                            storage_type&      storage,
+                                            BinaryFunction     compare_function)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
-        Key next_key = storage_.key[next_id];
-        bool b = next_id < flat_tid;
+        Key                        next_key = storage_.key[next_id];
+        bool                       b        = next_id < flat_tid;
         bool compare = compare_function(b ? key : next_key, b ? next_key : key);
-        bool swap = compare ^ dir;
+        bool                       swap     = compare ^ dir;
         if(swap)
         {
-            key = next_key;
+            key   = next_key;
             value = storage_.value[next_id];
         }
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void swap(Key (&key)[ItemsPerThread],
-              Value (&value)[ItemsPerThread],
-              const unsigned int flat_tid,
-              const unsigned int next_id,
-              const bool dir,
-              storage_type& storage,
-              BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void swap(Key (&key)[ItemsPerThread],
+                                            Value (&value)[ItemsPerThread],
+                                            const unsigned int flat_tid,
+                                            const unsigned int next_id,
+                                            const bool         dir,
+                                            storage_type&      storage,
+                                            BinaryFunction     compare_function)
     {
         storage_type_<Key, Value>& storage_ = storage.get();
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
-            Key next_key = storage_.key[item * BlockSize + next_id];
-            bool b = next_id < flat_tid;
-            bool compare = compare_function(b ? key[item] : next_key, b ? next_key : key[item]);
-            bool swap = compare ^ dir;
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
+            Key  next_key = storage_.key[item * BlockSize + next_id];
+            bool b        = next_id < flat_tid;
+            bool compare  = compare_function(b ? key[item] : next_key, b ? next_key : key[item]);
+            bool swap     = compare ^ dir;
             if(swap)
             {
                 key[item]   = next_key;
@@ -457,100 +447,94 @@ private:
         }
     }
 
-    template<
-        unsigned int Size,
-        class BinaryFunction,
-        class... KeyValue
-    >
+    template<unsigned int Size, class BinaryFunction, class... KeyValue>
     ROCPRIM_DEVICE ROCPRIM_INLINE
-    typename std::enable_if<(Size <= ::rocprim::device_warp_size())>::type
-    sort_power_two(const unsigned int flat_tid,
-                   storage_type& storage,
-                   BinaryFunction compare_function,
-                   KeyValue&... kv)
+        typename std::enable_if<(Size <= ::rocprim::device_warp_size())>::type
+        sort_power_two(const unsigned int flat_tid,
+                       storage_type&      storage,
+                       BinaryFunction     compare_function,
+                       KeyValue&... kv)
     {
-        (void) flat_tid;
-        (void) storage;
+        (void)flat_tid;
+        (void)storage;
 
         ::rocprim::warp_sort<Key, Size, Value> wsort;
         wsort.sort(kv..., compare_function);
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void warp_swap(Key& k, Value& v, int mask, bool dir, BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        warp_swap(Key& k, Value& v, int mask, bool dir, BinaryFunction compare_function)
     {
-        Key k1    = warp_shuffle_xor(k, mask);
+        Key  k1   = warp_swizzle_shuffle(k, mask);
         bool swap = compare_function(dir ? k : k1, dir ? k1 : k);
-        if (swap)
+        if(swap)
         {
             k = k1;
-            v = warp_shuffle_xor(v, mask);
+            v = warp_swizzle_shuffle(v, mask);
         }
     }
 
-    template <class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void warp_swap(Key (&k)[ItemsPerThread],
-                   Value (&v)[ItemsPerThread],
-                   int            mask,
-                   bool           dir,
-                   BinaryFunction compare_function)
+    template<class BinaryFunction>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void warp_swap(Key (&k)[ItemsPerThread],
+                                                 Value (&v)[ItemsPerThread],
+                                                 int            mask,
+                                                 bool           dir,
+                                                 BinaryFunction compare_function)
     {
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
-            Key k1    = warp_shuffle_xor(k[item], mask);
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
+            Key  k1   = warp_swizzle_shuffle(k[item], mask);
             bool swap = compare_function(dir ? k[item] : k1, dir ? k1 : k[item]);
-            if (swap)
+            if(swap)
             {
                 k[item] = k1;
-                v[item] = warp_shuffle_xor(v[item], mask);
+                v[item] = warp_swizzle_shuffle(v[item], mask);
             }
         }
     }
 
     template<class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void warp_swap(Key& k, int mask, bool dir, BinaryFunction compare_function)
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        warp_swap(Key& k, int mask, bool dir, BinaryFunction compare_function)
     {
-        Key k1    = warp_shuffle_xor(k, mask);
+        Key  k1   = warp_swizzle_shuffle(k, mask);
         bool swap = compare_function(dir ? k : k1, dir ? k1 : k);
-        if (swap)
+        if(swap)
         {
             k = k1;
         }
     }
 
-    template <class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void warp_swap(Key (&k)[ItemsPerThread], int mask, bool dir, BinaryFunction compare_function)
+    template<class BinaryFunction>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void
+        warp_swap(Key (&k)[ItemsPerThread], int mask, bool dir, BinaryFunction compare_function)
     {
         ROCPRIM_UNROLL
-        for(unsigned int item = 0; item < ItemsPerThread; ++item) {
-            Key k1    = warp_shuffle_xor(k[item], mask);
+        for(unsigned int item = 0; item < ItemsPerThread; ++item)
+        {
+            Key  k1   = warp_swizzle_shuffle(k[item], mask);
             bool swap = compare_function(dir ? k[item] : k1, dir ? k1 : k[item]);
-            if (swap)
+            if(swap)
             {
                 k[item] = k1;
             }
         }
     }
 
-    template <class BinaryFunction, unsigned int Items = ItemsPerThread, class... KeyValue>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    typename std::enable_if<(Items < 2)>::type
-    thread_merge(bool /*dir*/, BinaryFunction /*compare_function*/, KeyValue&... /*kv*/)
-    {
-    }
+    template<class BinaryFunction, unsigned int Items = ItemsPerThread, class... KeyValue>
+    ROCPRIM_DEVICE ROCPRIM_INLINE typename std::enable_if<(Items < 2)>::type
+        thread_merge(bool /*dir*/, BinaryFunction /*compare_function*/, KeyValue&... /*kv*/)
+    {}
 
-    template <class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void thread_swap(Key (&k)[ItemsPerThread],
-                     Value (&v)[ItemsPerThread],
-                     bool           dir,
-                     unsigned int   i,
-                     unsigned int   j,
-                     BinaryFunction compare_function)
+    template<class BinaryFunction>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void thread_swap(Key (&k)[ItemsPerThread],
+                                                   Value (&v)[ItemsPerThread],
+                                                   bool           dir,
+                                                   unsigned int   i,
+                                                   unsigned int   j,
+                                                   BinaryFunction compare_function)
     {
         if(compare_function(k[i], k[j]) == dir)
         {
@@ -562,13 +546,12 @@ private:
             v[j]         = v_temp;
         }
     }
-    template <class BinaryFunction>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void thread_swap(Key (&k)[ItemsPerThread],
-                     bool           dir,
-                     unsigned int   i,
-                     unsigned int   j,
-                     BinaryFunction compare_function)
+    template<class BinaryFunction>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void thread_swap(Key (&k)[ItemsPerThread],
+                                                   bool           dir,
+                                                   unsigned int   i,
+                                                   unsigned int   j,
+                                                   BinaryFunction compare_function)
     {
         if(compare_function(k[i], k[j]) == dir)
         {
@@ -578,19 +561,15 @@ private:
         }
     }
 
-    template <class BinaryFunction, class... KeyValue>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    void thread_shuffle(unsigned int offset, bool dir, BinaryFunction compare_function, KeyValue&... kv)
+    template<class BinaryFunction, class... KeyValue>
+    ROCPRIM_DEVICE ROCPRIM_INLINE void thread_shuffle(unsigned int   offset,
+                                                      bool           dir,
+                                                      BinaryFunction compare_function,
+                                                      KeyValue&... kv)
     {
         ROCPRIM_UNROLL
         for(unsigned base = 0; base < ItemsPerThread; base += 2 * offset)
         {
-            ROCPRIM_UNROLL
-// Workaround to prevent the compiler thinking this is a 'Parallel Loop' on clang 15
-// because it leads to invalid code generation with `T` = `char` and `ItemsPerthread` = 4
-#if defined(__clang_major__) && __clang_major__ >= 15
-    #pragma clang loop vectorize(disable)
-#endif
             for(unsigned i = 0; i < offset; ++i)
             {
                 thread_swap(kv..., dir, base + i, base + i + offset, compare_function);
@@ -598,16 +577,15 @@ private:
         }
     }
 
-    template <class BinaryFunction, unsigned int Items = ItemsPerThread, class... KeyValue>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    typename std::enable_if<!(Items < 2)>::type
-    thread_merge(bool dir, BinaryFunction compare_function, KeyValue&... kv)
+    template<class BinaryFunction, unsigned int Items = ItemsPerThread, class... KeyValue>
+    ROCPRIM_DEVICE ROCPRIM_INLINE typename std::enable_if<!(Items < 2)>::type
+        thread_merge(bool dir, BinaryFunction compare_function, KeyValue&... kv)
     {
         ROCPRIM_UNROLL
         for(unsigned int k = ItemsPerThread / 2; k > 0; k /= 2)
         {
             thread_shuffle(k, dir, compare_function, kv...);
-        }    
+        }
     }
 
     /// Bitonic sort.
@@ -622,14 +600,14 @@ private:
     {
         const auto warp_id_is_even = ((flat_tid / ::rocprim::device_warp_size()) % 2) == 0;
         ::rocprim::warp_sort<Key, ::rocprim::device_warp_size(), Value> wsort;
-        auto compare_function2 =
-            [compare_function, warp_id_is_even](const Key& a, const Key& b) mutable -> bool
-            {
-                auto r = compare_function(a, b);
-                if(warp_id_is_even)
-                    return r;
-                return !r;
-            };
+        auto                                                            compare_function2
+            = [compare_function, warp_id_is_even](const Key& a, const Key& b) mutable -> bool
+        {
+            auto r = compare_function(a, b);
+            if(warp_id_is_even)
+                return r;
+            return !r;
+        };
         wsort.sort(kv..., compare_function2);
 
         ROCPRIM_UNROLL
@@ -644,11 +622,14 @@ private:
                 ::rocprim::syncthreads();
             }
 
+            const unsigned int     id = detail::logical_lane_id<::rocprim::device_warp_size()>();
+            constexpr unsigned int s  = ::rocprim::device_warp_size() / 2;
+
             ROCPRIM_UNROLL
-            for(unsigned int k = ::rocprim::device_warp_size() / 2; k > 0;  k /= 2)
+            for(unsigned int k = s; k > 0; k /= 2)
             {
-                const bool length_even = ((detail::logical_lane_id<::rocprim::device_warp_size()>() / k ) % 2 ) == 0;
-                const bool local_dir = length_even ? dir : !dir;
+                const bool length_even = ((id / k) % 2) == 0;
+                const bool local_dir   = length_even ? dir : !dir;
                 warp_swap(kv..., k, local_dir, compare_function);
             }
             thread_merge(dir, compare_function, kv...);
